#
#

class Events(object):
    #
    @staticmethod
    def to_event_string(event):
        eventStr = event
        # TODO
        # if (typeof event === 'string') {
        #     eventStr = event
        # }
        # else if (typeof event === 'object' && event.event) {
        #     eventStr = event.event
        # }
        # else 
        #     throw new Error ('Unknown event object: should be a string or object with event string')
        return eventStr

    @classmethod
    def to_consume_event(cls, event):
        eventStr = cls.to_event_string(event)
        return 'CONSUME-' + eventStr

    @staticmethod
    def to_ondisconnect_event(id):
        return 'DISCONNECT-' + id

    @classmethod
    def to_onunsubscribe_event(cls, event, id):
        eventStr = cls.to_event_string(event)
        return 'UNSUBSCRIBE-' + eventStr + '-' + id

    @staticmethod
    def to_onsubscribe_event(id):
        return 'SUBSCRIBE-TO' + ("-" + id if (id is not None) else "")


#
#
from .constants import Constants
from .publisher import Publisher
from .subscriber import Subscriber
from .socket import Socket
from .events import Events
from .logger import Logger

class MessageQueue(object):

    def __init__(self, host=None, port=None, protocol=None):
        # The SocketIO instance
        if not host is None:
            self.host = host

        if port is not None:
            self.port = port

    @staticmethod
    def createSocket():
        mySocket = Socket()

        return mySocket

    #/**
     #* private function
     #*/
    @staticmethod
    def createConsumerPrivate(name, host=None, port=None, protocol=None): 
        consumer = Subscriber(name, host, port, protocol)
        return consumer


    # /**
    #  * Create a consumer
    #  */
    @classmethod
    def createConsumer(cls, name, host=None, port=None, protocol=None):
        return cls.createConsumerPrivate(name, host, port, protocol)

    # /**
    #  * Alias of createConsumer
    #  */

    def createSubscriber(self, name, host=None, port=None, protocol=None):
        return self.createConsumer(name, host, port, protocol)

    @staticmethod
    def createProducerPrivate (name, eventDefault=None, host=None, port=None, protocol=None):
        producer = Publisher(name, eventDefault, host, port, protocol)
        return producer
     
    # /**
    #  * Create a producer
    #  */
    def createProducer (self, name, eventDefault=None, host=None, port=None, protocol=None):
        return  self.createProducerPrivate(name, eventDefault, host, port, protocol)

    def createPublisher (self, name, eventDefault=None, host=None, port=None, protocol=None):
        return self.createProducer(name, eventDefault, host, port, protocol)


#
# 
from .constants import Constants
from .logger import Logger
from .base import Base
 
from socketIO_client import SocketIO, BaseNamespace

import uuid

class SocketListener(BaseNamespace):

    def on_connect(self):
        print('[Connected]')

    def on_reconnect(self):
        print('[Reconnected]')

    def on_disconnect(self):
        print('[Disconnected]')


class Socket(Base):
    
    #
    def __init__(self, host=None, port=None, protocol=None):
        self.type = 'RAW'

        self.autoreconnect = True
        if (host is None):
            self.host = 'localhost'
        else:
            self.host = host

        if (port is None):
            self.port = Constants.DEFAULT_PORT
        else:
            self.port = port

        if (protocol is not None):
            self.protocol = protocol

        self.socket = None
        self.connected = False
        self.id = str(uuid.uuid4())
        self.name = Constants.ANONYMOUS
        self.alias = None

        #
        self.on_connect_listeners = []
        self.on_error_listener = None
        self.on_event_func_list = None

    def __apply_on_events (self):
        if (self.on_event_func_list is not None and len(self.on_event_func_list) > 0):
            map(lambda func : func(), self.on_event_func_list)
            del self.on_event_func_list
            self.on_event_func_list = None

    def send_identification_info(self):
        self.send_message(self.type, {'name': self.name, 'id':self.get_id()})

    def on_connect(self):
        Logger.log("connected to message queue server")

        self.connected = True
        self.socket.on('ERROR', self.__on_error__)

        self.send_identification_info()

        i = 0
        while (i < len(self.on_connect_listeners)):
            listener = self.on_connect_listeners[i]
            listener()
            i += 1

    def on_disconnect(self):
        self.connected = False
        #Logger.debug('disconnect')
        Logger.log("Socket (" + self.get_id() + ") is disconnected")

    def on_reconnect(self):
        Logger.debug('reconnect')

    def on_error(self):
        Logger.error('oops, something wrong.')
        
    #
    # On TYO-MQ ERROR MESSAGE
    #
    def __on_error__(self, msg):
        if (self.on_error_listener is not None):
            self.on_error_listener()
        else:
            Logger.error(msg)

    #
    #
    #
    def connect(self, duration=-1, callback=None):
        # Example
        # with SocketIO(self.host, self.port, SocketListener) as socketIO:
        #     socketIO.emit('aaa')
        #     socketIO.wait(seconds=1)
        self.socket = SocketIO(self.host, self.port, SocketListener)
        self.socket.on('connect', self.on_connect if callback is None  else callback)
        self.socket.on('disconnect', self.on_disconnect)
        self.socket.on('reconnect', self.on_reconnect)

        if duration == -1 :
            self.socket.wait()
        else:
            self.socket.wait(seconds=duration)


    def get_id(self):
        return self.id

    def add_on_connect_listener(self, listener):
        self.on_connect_listeners.append(listener)

    def disconnect(self):
        if (self.socket is not None and self.connected):
            self.socket.disconnect()

    def on(self, event, callback):
        if (self.socket is None):
            #raise Exception("Socket is not created yet")
            if self.on_event_func_list is None:
                self.on_event_func_list = []
                futureFunc = lambda : self.__apply_on_events()
                self.add_on_connect_listener(futureFunc)

            self.on_event_func_list.append(callback)
        else: 
            self.socket.on(event, callback)

    def send_message(self, event, msg):
        if (self.socket is None):
            raise Exception("Socket isn't ininitalized yet")

        if (self.socket.connected is False):
            futureFunc = lambda event, msg: self.socket.emit(event, msg)

            if (self.autoreconnect is True):
                    self.connect(-1, futureFunc)
            else:
                raise Exception("Socket is created but not connected")
        else:
            self.socket.emit(event, msg)






#
#
from .subscriber import Subscriber
from .logger import Logger
from .constants import Constants
from .events import Events

#
import json

class Publisher(Subscriber):
    def __init__(self, name, eventDefault=None, host=None, port=None, protocol=None):
        super(Publisher, self).__init__(name, host, port, protocol)

        self.type = 'PRODUCER'
        self.eventDefault = eventDefault if eventDefault is not None else Constants.EVENT_DEFAULT
        self.on_subscription_listener = None
        self.subscribers = {}

        # // Initialisation
        futureFunc = lambda : self.set_on_subscription_listener()
        self.add_on_connect_listener(futureFunc)

        #
        Logger.debug("creating producer: " + self.name)

    def broadcast (self, data, event=None):
        self.produce(data, event, Constants.METHOD_BROADCAST)
        
    def produce (self, data, event=None, method=None) :   
        if (data is None):
             raise Exception("data can't be null")
        
        if (event is None):
            if (self.eventDefault is None):
                raise Exception("please specifiy event")
            else:
                 event = self.eventDefault   

        message = {"event":event, "message":data, "from":self.name, "method":method}
        self.send_message('PRODUCE', message)     

    # /**
    #  * On Subscribe
    #  */
    def __on_subscription (self, data) :
        Logger.log("Received subscription information: " + json.dumps(data))

        self.subscribers[data["id"]] = data

        # // further listener
        if (self.on_subscription_listener is not None):
            self.on_subscription_listener(data)

    def set_on_subscription_listener (self) :
        event = Events.to_onsubscribe_event(self.get_id())
        self.on(event, self.__on_subscription)

    # /**
    #  * On Lost connections with subscriber(s)
    #  */
    def __on_lost_subscriber (self, callback, data) :
        Logger.log("Lost subscriber's connection")
        if (callback is not None):
            callback(data)

    def set_on_subscriber_lost_listener (self, callback) :
        event = Events.to_ondisconnect_event(self.get_id())
        futureFunc = lambda data : (lambda data, cb=callback : self.__on_lost_subscriber(cb, data))(data)
        self.on(event, futureFunc)

    def on_subscriber_lost (self, callback) : 
        self.set_on_subscriber_lost_listener(callback)

    # /**
    #  * On Unsubsribe
    #  */
    def __on_unsubscribed (self, callback, data) : 
        if callback is not None:
            callback(data)

    def set_on_unsubscribed_listener (self, event, callback) :
        event = Events.to_onunsubscribe_event(event, self.get_id())
        futureFunc = lambda data : (lambda data, cb=callback: self.__on_unsubscribed(cb, data))(data)
        self.on(event, futureFunc)

    def on_unsubscribed (self, event, callback) :
        self.set_on_unsubscribed_listener(event, callback)



from .socket import Socket
from .logger import Logger
from .events import Events
from .constants import Constants

#
import json, sys

class Subscriber(Socket):

    #
    def __init__(self, name=None, host=None, port=None, protocol=None):
        super(Subscriber, self).__init__(host, port, protocol)

        self.type = 'CONSUMER'
        self.name = name if name is not None else Constants.ANONYMOUS
        self.consumes = None
        self.subscriptions = None
        Logger.debug("creating subscriber: " + self.name)

    # def send_identification_info(self):
    #     self.send_message(, {'name': self.name})

    def __apply_subscritptions(self):
        if (len(self.subscriptions) > 0):
            map(lambda func : func(), self.subscriptions)
            del self.subscriptions
            self.subscriptions = None

    def __trigger_consume_event(self, obj, eventStr, callback) :
        #if obj["event"] == eventStr :
        callback(obj["message"])

    #
    # For debug
    # 
    def __debug_on_message(self, event, message):
        Logger.debug('received message', event, message)
        try:
            func = self.consumes[event]
            self.__trigger_consume_event(message, event, func)
        except :
            Logger.error("Ooops, something wrong", sys.exc_info()[0])
            raise
    #     Logger.debug(event, ":", json.dumps(message))
    #     callback(message)

    # def __debug_on_message(self, *args):
    #     Logger.debug('received message', args)

    def __subscribe_internal(self, who, event, onConsumeCallback):    
            eventStr = None
            if (event is not None):
                eventStr = Events.to_event_string(event)
            else:
                eventStr = who + "-ALL"

            # /**
            #  * @todo
            #  * 
            #  * deal with the ALL events later
            #  */
    
            sendSubscriptionMessage = lambda event=eventStr, who=who, name=self.name : self.send_message('SUBSCRIBE', {"event":event, "producer":who, "consumer":name})

            # // On Connect Message will be trigger by system
            if (self.connected):
                sendSubscriptionMessage()
            else:
                if self.subscriptions is None:
                    self.subscriptions = []
                    futureFunc = lambda : self.__apply_subscritptions()
                    self.add_on_connect_listener(futureFunc)

                self.subscriptions.append(sendSubscriptionMessage)
            # // the connection should be ready before we subscribe the message
            # // self.on('connect', function ()  {
            # //     sendSubscriptionMessage()
            # // })
    
            if (self.consumes is None):
                self.consumes = {}
    
            consumeEventStr = Events.to_consume_event(eventStr)
            self.consumes[consumeEventStr] = onConsumeCallback #lambda message, fromWhom : onConsumeCallback(message, fromWhom)
            #lambda obj : lambda obj, event=eventStr, callback=onConsumeCallback : self.__trigger_consume_event(obj, event, callback)

            #futureFunc = lambda data : (lambda data, event=consumeEventStr: self.consumes[event](data))(data)
            #futureFunc = lambda data, eventStr=consumeEventStr : self.consumes[eventStr](data)
            #DEBUG
            Logger.debug("setting on event: " + consumeEventStr)
            #self.on(consumeEventStr, self.__debug_on_message)
            #futureFunc = lambda data : self.__debug_on_message(data)
            futureFunc = lambda data, event=consumeEventStr : self.__debug_on_message(event, data)
            self.on(consumeEventStr, futureFunc)

    def resubscribeWhenReconnect (self, who, event, onConsumeCallback, reSubscribe=True):

        resubscribeListener = lambda who=who, eventStr=event, callback=onConsumeCallback: self.__subscribe_internal(who, eventStr, callback)

        self.__subscribe_internal(who, event, onConsumeCallback)

        if (reSubscribe is True):
            self.add_on_connect_listener(resubscribeListener)

    # /**
    #  * Subscribe message
    #  * 
    #  * If an event name is not provided, then we subscribe all the messages from the producer
    #  */

    def subscribe (self, who, event, onConsumeCallback, reconcect=True):
        self.resubscribeWhenReconnect(who, event, onConsumeCallback, reconcect)

    # /**
    #  * Subscribe only once, if the connection is gone, let it be
    #  */

    def subscribeOnce (self, who, event, onConsumeCallback):
        self.subscribe(who, event, onConsumeCallback, False)

    # /**
    #  * Subscribe all events with this name whatever providers are publishing
    #  */
    def subscribeAll (self, event, onConsumeCallback):
        self.subscribe(Constants.ALL_PUBLISHERS, event, onConsumeCallback, True)


#
#

class Constants:

    ANONYMOUS = 'ANONYMOUS'
    EVENT_DEFAULT = 'tyo-mq-mt-default'
    SYSTEM = 'TYO-MQ-SYSTEM'
    ALL_PUBLISHERS = 'TYO-MQ-ALL'

    METHOD_BROADCAST = 'broadcast'

    # SERVER DEFAULTS
    DEFAULT_PORT = 17352


#
#
from __future__ import print_function
import sys

class Logger:
    instance = None

    def __init__(self, level):
        self.level = level

    @staticmethod
    def log(*args, **kwargs):
        print (args, kwargs)

    @staticmethod
    def error(*args, **kwargs):
        print(*args, file=sys.stderr, **kwargs)

    @staticmethod
    def warn(*args, **kwargs):
        print (args, kwargs)

    @staticmethod
    def verbose(*args, **kwargs):
        print (args, kwargs)

    @staticmethod
    def debug(*args, **kwargs):
        print (args, kwargs)

    @staticmethod
    def info(*args, **kwargs):
        print (args, kwargs)